from statistics import mean
import logging
from ....models.hourselection.cautionhourtype import CautionHourType, MAX_HOURS

_LOGGER = logging.getLogger(__name__)


def normalize_prices(prices:list) -> list:
    min_price = min(prices)
    c = 0
    if min_price <= 0:
        c = abs(min_price) + 0.01
    ret = []
    for p in prices:
        pp = p+c
        divider = min_price if min_price > 0 else c
        ret.append(round(pp-divider,3))
    return ret

def get_offset_dict(normalized_hourdict: dict):
        ret = {}
        _prices = [p-min(normalized_hourdict.values()) for p in normalized_hourdict.values()]
        average_val = mean(_prices)
        for i in range(0,len(_prices)):
            try:
                ret[i] = round((_prices[i]/average_val) - 1,2)
            except:
                ret[i] = 1
        return ret

def create_cautions(
        hourdict: dict, 
        normalized_hourdict: dict, 
        cautionhour_type: CautionHourType, 
        range_start: int = 0,
        adjusted_average:float = None, 
        blocknocturnal:bool = False
        ) -> dict:
    """Rank the normalized pricelist to find out which are going to become non- or caution-hours"""
    
    _adj_avg = adjusted_average
    if not isinstance(adjusted_average, (float,int)):
        _adj_avg = mean(hourdict.values()) 

    adj_average_norm = _adj_avg * (mean(normalized_hourdict.values())/mean(hourdict.values()))
    cautions = [h for h in normalized_hourdict if normalized_hourdict[h] > (adj_average_norm * 0.7)]
    cautions_dict = _cap_pricelist_available_hours(cautions, normalized_hourdict, cautionhour_type, blocknocturnal, range_start)
    maxval = max(hourdict.values())
    ret = {}
    
    try:
        for k, v in cautions_dict.items():
            ret[k] = {
                "val": hourdict[k], 
                "permax": round(hourdict[k] / maxval,2), 
                "force_non": v
                }
    except IndexError as e:
        _LOGGER.error(f"Error on creating the cautions_dict: {e}")

    if blocknocturnal:
        return ret
    else:
        return _discard_excessive_hours(ret)

def _cap_pricelist_available_hours(cautions: list, normalized_hourdict:dict, cautionhour_type: CautionHourType, blocknocturnal:bool, range_start: int) -> dict:
    _demand = max(len(normalized_hourdict.keys()) - MAX_HOURS.get(cautionhour_type),0)
    ret = {c: False for c in cautions} if _demand == 0 else {c: True for c in cautions}
    hours_sorted = [k for k, v in sorted(normalized_hourdict.items(), key=lambda item: item[1])]
    iterations = 0
    while len(cautions) < _demand and iterations < len(hours_sorted)*2:
        iterations+=1
        idx = max(range_start-1,0)
        if len(cautions) > 0:
            idx = hours_sorted.index(cautions[-1])
        
        if idx <= len(hours_sorted) -1:
            try:
                while idx+1 < len(hours_sorted):
                    idx += 1
                    next = hours_sorted[idx]
                    if next not in cautions:
                        cautions.append(next)
                        ret[next] = True
                        break
            except IndexError as e:
                raise IndexError(f"error on first. idx:{idx}")
        if idx >= 0:
            try:
                while idx-1 >= 0:
                    idx -= 1
                    prev = hours_sorted[idx]
                    if prev not in cautions:
                        cautions.append(prev)
                        ret[prev] = True
                        break
            except IndexError as e:
                raise IndexError(f"error on second. idx:{idx}")

    try:
        for i in get_nocturnal_stop(blocknocturnal, range_start):
            if i not in cautions:
                ret[i] = True
    except IndexError as e:
        _LOGGER.error(f"error on looping nocturnal stop: {e}")
    return _sort_by_key(ret)

def _sort_by_key(input: dict) -> dict:
    _keys = list(input.keys())
    _keys.sort()
    return {i: input[i] for i in _keys}

def get_nocturnal_stop(blocknocturnal: bool = False, range_start: int = 0) -> list:
    _base = [23,0,1,2,3,4,5,6]
    if blocknocturnal:
        return _transform_range(range_start, _base)
    return []


def _transform_range(range_start, base):
    if range_start == 0:
            return base
    ret = []
    for b in base:
        if b >= range_start:
            ret.append(b-range_start)
        else:
            ret.append(24-range_start+b)
    return ret


def _discard_excessive_hours(hours: dict):
    """There should always be at least four regular hours before absolute_top_price kicks in."""
    while len(hours) >= 20:
        to_pop = dict(sorted(hours.items(), key=lambda item: item[1]['val']))    
        hours.pop(list(to_pop.keys())[0])
    return hours


def should_be_cautionhour(price_item, prices, peak, cautionhour_type) -> bool:
    first = any([
                float(price_item["permax"]) <= cautionhour_type,
                float(price_item["val"]) <= (sum(prices)/len(prices))
            ])
    second = (peak > 0 and peak*price_item["permax"] > 1) or peak == 0
    return all([first, second])


def set_charge_allowance(price_input, cautionhour_type) -> float:
    return round(abs(price_input - 1), 2) * ALLOWANCE_SCHEMA[cautionhour_type]

ALLOWANCE_SCHEMA = {
    CautionHourType.get_num_value(CautionHourType.SUAVE): 1.15,
    CautionHourType.get_num_value(CautionHourType.INTERMEDIATE): 1.05,
    CautionHourType.get_num_value(CautionHourType.AGGRESSIVE): 1,
    CautionHourType.get_num_value(CautionHourType.SCROOGE): 1
}


import logging
import time
from datetime import datetime

_LOGGER = logging.getLogger(__name__)

class KillSwitch:
    def __init__(
        self, 
        sensor: str, 
        update_interval:int = 120, 
        grace_interval:int = 120
        ):
        self._update_interval: int = update_interval
        self._grace_interval: int = grace_interval
        self._bound_sensor: str = sensor
        self._last_update: int = time.time()
        self._grace_period: bool = False
        self._grace_period_start: int = 0
        self._is_dead: bool = False

    @property
    def total_timer(self) -> int:
        return sum(self._update_interval, self._grace_interval)

    @property
    def is_dead(self) -> bool:
        return self._is_dead

    @property
    def check(self):
        if self._grace_period:
            if time.time() - self._grace_period_start > self._grace_interval:
                self._is_dead = True
                #self.__init__(self._bound_sensor, self._update_interval, self._grace_interval)
        elif time.time() - self._last_update > self._update_interval:
            _LOGGER.warning(f"{datetime.now()} - Caution! The sensor {self._bound_sensor} has not been updated for {self._update_interval} seconds. In {self._grace_interval} seconds the charger will be stopped unless this sensor comes alive again.")
            self._grace_period = True
            self._grace_period_start = time.time()

    def update(self):
        self._last_update = time.time()
        self._grace_period = False
        self._is_dead = False
        self._grace_period_start = 0

    async def async_update(self):
        self._last_update = time.time()
        self._grace_period = False
        self._is_dead = False
        self._grace_period_start = 0

    


# k = KillSwitch("magnus_sensor", 3, 6)
# k.update()
# time.sleep(1)
# k.check()
# k.update()
# time.sleep(4)
# k.check()
# time.sleep(5)
# k.check()
# time.sleep(2)
# k.check()



# import logging
# import time
# from datetime import datetime
# from ...models.chargecontroller_states import ChargeControllerStates
# from abc import abstractmethod
# from typing import List, Dict

# _LOGGER = logging.getLogger(__name__)


# class ChargeControllerBase:  
#     def __init__(
#         self,
#         charger_state_translation:Dict[ChargeControllerStates,List[str]],
#         non_hours:List[int] = [],
#         timeout:int = 300
#     ):
#         self._charger_state_translation = self._check_charger_states(charger_state_translation)
#         self.done_timeout: int = timeout
#         self._latest_charger_start: float = time.time()
#         self._non_hours = non_hours

#     @property
#     def _is_timeout(self) -> bool:
#         return time.time() - self.latest_charger_start > self._done_timeout

#     @property
#     def latest_charger_start(self) -> float:
#         return self._latestchargerstart

#     @latest_charger_start.setter
#     def latest_charger_start(self, val):
#         self._latestchargerstart = val

#     @property
#     @abstractmethod
#     def status(self) -> str:
#         if self._hub.is_initialized is False:
#             return "Hub not ready. Check logs!"
#         if self._hub.is_initialized is True:
#             if self._chargecontroller_initalized is False:
#                 self._chargecontroller_initalized = True
#                 _LOGGER.debug("Chargecontroller is initialized and ready to work!")
#         if self._hub.options.charger.charger_is_outlet is True:
#             ret = self._get_status_outlet()
#         else:
#             ret = self._get_status()
#         if ret == ChargeControllerStates.Error:
#             msg = f"Chargecontroller returned faulty state. Charger reported {self._hub.chargerobject.value.lower()} as state."
#             _LOGGER.error(msg)
#         return ret.name

#     @property
#     @abstractmethod
#     def charger_state(self) -> ChargeControllerStates:
#         """
#         Check if you are below the starting threshold of charger
#         :return: returns enum ChargeControllerStates
#         """
#         pass

#     @property
#     @abstractmethod
#     def charger_done(self) -> bool:
#         """
        
#         :return: returns bool
#         """
#         pass

#     @charger_done.setter
#     @abstractmethod
#     def charger_done(self, val) -> None:
#         """
        
#         :return: returns None
#         """
#         pass

#     @property
#     @abstractmethod
#     def charger_enabled(self) -> bool:
#         """
        
#         :return: returns bool
#         """
#         pass

#     @property
#     @abstractmethod
#     def free_charge(self) -> bool:
#         """
        
#         :return: returns bool
#         """
#         pass

#     @abstractmethod
#     def _get_status_charging(self) -> ChargeControllerStates:
#         """
        
#         :return: returns enum ChargeControllerStates
#         """
#         pass

#     @abstractmethod
#     def _get_status_connected(self) -> ChargeControllerStates:
#         """
        
#         :return: returns enum ChargeControllerStates
#         """
#         pass

#     def _get_status_outlet(self):
#         ret = ChargeControllerStates.Error
#         update_timer = False
#         free_charge = self._hub.locale.data.free_charge(self._hub.locale.data)

#         if self._hub.charger_enabled.value is False:
#             update_timer = True
#             ret = ChargeControllerStates.Disabled
#         elif self._hub.charger_done.value is True:
#             ret = ChargeControllerStates.Done
#         elif datetime.now().hour in self._hub.non_hours and free_charge is False and self._hub.timer.is_override is False:
#             update_timer = True
#             ret = ChargeControllerStates.Stop
#         elif self._hub.chargertype.entities.powerswitch == "on" and self._hub.chargertype.entities.powermeter < 1:
#             ret = self._get_status_connected()
#             update_timer = (ret == ChargeControllerStates.Stop)
#         else:
#             ret = self._get_status_charging()
#             update_timer = True

#         if update_timer is True:
#             self.update_latestchargerstart()
#         return ret

#     def _get_status(self):
#         ret = ChargeControllerStates.Error
#         update_timer = False
#         charger_state = self._hub.chargerobject.value.lower()
#         free_charge = self._hub.locale.data.free_charge(self._hub.locale.data)

#         if self._hub.charger_enabled.value is False:
#             update_timer = True
#             ret = ChargeControllerStates.Disabled
#         elif charger_state in self._hub.chargertype.chargerstates[ChargeControllerStates.Done]:
#             self._hub.charger_done.value = True
#             ret = ChargeControllerStates.Done
#         elif charger_state in self._hub.chargertype.chargerstates[ChargeControllerStates.Idle]:
#             update_timer = True
#             ret = ChargeControllerStates.Idle
#             if self._hub.charger_done.value is True:
#                 self._hub.charger_done.value = False
#         elif charger_state not in self._hub.chargertype.chargerstates[ChargeControllerStates.Idle] and self._hub.charger_done.value is True:
#             ret = ChargeControllerStates.Done
#         elif datetime.now().hour in self._hub.non_hours and free_charge is False and self._hub.timer.is_override is False:
#             update_timer = True
#             ret = ChargeControllerStates.Stop
#         elif charger_state in self._hub.chargertype.chargerstates[ChargeControllerStates.Connected]:
#             ret = self._get_status_connected(charger_state)
#             update_timer = (ret == ChargeControllerStates.Stop)
#         elif charger_state in self._hub.chargertype.chargerstates[ChargeControllerStates.Charging]:
#             ret = self._get_status_charging()
#             update_timer = True

#         if update_timer is True:
#             self.update_latestchargerstart()
#         return ret

#     def _is_done(self, charger_state) -> bool:
#         if len(self._hub.chargertype.chargerstates[ChargeControllerStates.Done]):
#             return charger_state in self._hub.chargertype.chargerstates[ChargeControllerStates.Done]
#         return time.time() - self.latest_charger_start > self.done_timeout

#     def update_latestchargerstart(self):
#         self.latest_charger_start = time.time()

#     def _below_start_threshold(
#             self,
#             predicted_energy: float,
#             current_peak: float,
#             threshold_start: float
#     ) -> bool:
#         """
#         Check if you are below the starting threshold of charger
#         :param predicted_energy: predicted energy for this hour in kWh
#         :param current_peak: currently set energy peak, in kWh
#         :param threshold_start: the allowed peak percentage to start. 0-1
#         :return: returns bool
#         """
#         return (predicted_energy * 1000) < ((current_peak * 1000) * threshold_start)

#     def _above_stop_threshold(
#             self,
#             predicted_energy: float,
#             current_peak: float,
#             threshold_stop: float
#     ) -> bool:
#         """
#         Check if you are above the stopping threshold of charger
#         :param predicted_energy: predicted energy for this hour in kWh
#         :param current_peak: currently set energy peak, in kWh
#         :param threshold_stop: the limit peak percentage to stop. 0-1
#         :return: returns bool
#         """
#         return (predicted_energy * 1000) > ((current_peak * 1000) * threshold_stop)

#     def _check_charger_states(self, input:Dict[ChargeControllerStates,List[str]]) -> Dict[ChargeControllerStates,List[str]]:
#         if len(input) == 0:
#             raise AssertionError
#         return input


    

    



# from ...models.chargecontroller_states import ChargeControllerStates
# from .chargecontrollerbase import ChargeControllerBase
# #from custom_components.peaqev.peaqservice.chargecontroller.chargecontrollerbase import ChargeControllerBase


# class ChargeController(ChargeControllerBase):
#     def __init__(self, hub):
#         super().__init__(charger_state_translation=self.self._hub.chargertype.chargerstates)

#     @property
#     def below_startthreshold(self) -> bool:
#         return self._below_start_threshold(
#             predicted_energy=self._hub.prediction.predictedenergy,
#             current_peak=self._hub.current_peak_dynamic,
#             threshold_start=self._hub.threshold.start/100
#         )

#     @property
#     def above_stopthreshold(self) -> bool:
#         return self._above_stop_threshold(
#             predicted_energy=self._hub.prediction.predictedenergy,
#             current_peak=self._hub.current_peak_dynamic,
#             threshold_stop=self._hub.threshold.stop/100
#         )

#     def _get_status_charging(self) -> ChargeControllerStates:
#         if self.above_stopthreshold and self._hub.totalhourlyenergy.value > 0 and self._hub.locale.data.free_charge(self._hub.locale.data) is False:
#             return ChargeControllerStates.Stop
#         else:
#             return ChargeControllerStates.Start

#     def _get_status_connected(self, charger_state=None) -> ChargeControllerStates:
#         if charger_state is not None and self._hub.sensors.carpowersensor.value < 1 and self._is_done(charger_state):
#             return ChargeControllerStates.Done
#         else:
#             if (self.below_startthreshold and self._hub.sensors.totalhourlyenergy.value != 0) or self._hub.sensors.locale.data.free_charge(self._hub.sensors.locale.data) is True:
#                 return ChargeControllerStates.Start
#             else:
#                 return ChargeControllerStates.Stop


from datetime import datetime, date, time
import math
from ...models.hourselection.hourselectionmodels import HourSelectionOptions
from .schedule_session import ScheduleSession
from ...models.chargecontroller_states import ChargeControllerStates




class Scheduler:
    """This class obj is what constitutes a running scheduler."""
    def __init__(self, options:HourSelectionOptions = None):
        self.model = ScheduleSession(hourselection_options=options)
        self.active = False

    @property
    def scheduler_active(self) -> bool:
        if self.active is False:
            return False
        return self.model.departuretime > datetime.now() or self.model.remaining_charge > 0

    def create(
        self, 
        desired_charge:float, 
        departuretime:datetime, 
        starttime:datetime=datetime.now(),
        override_settings = False
        ):
        if self.scheduler_active:
            self._cancel()
        self.model.departuretime = departuretime
        self.model.starttime = starttime
        self.model.remaining_charge = desired_charge
        self.model._override_settings = override_settings

    def _update(
        self,
        avg24:float,
        peak:float,
        charged_amount:float = None,
        prices:list = None,
        prices_tomorrow:list = None,
        mockdt:datetime = datetime.now()
        ):
        """calculate based on the pricing of hours, current peak and the avg24hr energy consumption"""
        self.model.MOCKDT = mockdt
        self.active = True
        if self.model.remaining_charge <= 0 or self.model.departuretime <= mockdt:
            return self._cancel()
        charge_per_hour = peak - (avg24/1000)
        if charge_per_hour <= 0:
            raise Exception
        
        self.model.remaining_charge -= charged_amount if charged_amount is not None else 0
        self.model.hours_price = [prices, prices_tomorrow]
        cheapest = self._sort_pricelist()
        self.model.hours_charge = self._get_charge_hours(
            cheapest_hours=cheapest, 
            charge_per_hour=charge_per_hour,
            peak=peak
            )

    def _cancel(self):
        self.active = False
        self.model.departuretime = datetime.min
        self.model.starttime = datetime.min
        self.model.remaining_charge = 0
    
    def _sort_pricelist(self) -> dict:
        if self.model._override_settings is False and self.model.hourselection_options is not None:
            return self._filter_pricelist()    
        return dict(sorted(self.model.hours_price.items(), key=lambda item: item[1]))
    
    def _filter_pricelist(self) -> dict:
        filtered = {key:value for (key,value) in self.model.hours_price.items() if value <= self.model.hourselection_options.absolute_top_price}
        ret= dict(sorted(filtered.items(), key=lambda item: item[1]))
        return ret

    def _get_charge_hours(
        self, 
        cheapest_hours:dict, 
        charge_per_hour:float,
        peak:float
        ) -> dict:
        remainder = self.model.remaining_charge
        chargehours = dict()
        for c in cheapest_hours.keys():
            if remainder <= 0:
                break
            if remainder > charge_per_hour:
                chargehours[c] = 1
            elif 0 < remainder < charge_per_hour:
                chargehours[c] = math.ceil((remainder/peak)*10)/10
            remainder -= charge_per_hour
        return chargehours


class SchedulerFacade(Scheduler):
    def __init__(self, hub, options):
        self._hub = hub
        super().__init__(options)
        self.schedule_created = False

    def create_schedule(self, charge_amount: float, departure_time: datetime, schedule_starttime: datetime, override_settings: bool = False):
        if not self.scheduler_active:
            self.create(charge_amount, departure_time, schedule_starttime, override_settings)
        self.schedule_created = True

    async def async_create_schedule(self, charge_amount: float, departure_time: datetime, schedule_starttime: datetime, override_settings: bool = False):
        if not self.scheduler_active:
            await self.hub.state_machine.async_add_executor_job(self.create,charge_amount, departure_time, schedule_starttime, override_settings)
        self.schedule_created = True

    def update(self):
        self._update(
            avg24=self._hub.sensors.powersensormovingaverage24.value,
            peak=self._hub.current_peak_dynamic,
            charged_amount=self._hub.charger.session.session_energy,
            prices=self._hub.hours.prices,
            prices_tomorrow=self._hub.hours.prices_tomorrow
        )
        self.check_states()

    async def async_update(self):
        await self.hub.state_machine.async_add_executor_job(self._update, 
                                                            self._hub.sensors.powersensormovingaverage24.value,
                                                            self._hub.current_peak_dynamic,
                                                            self._hub.charger.session.session_energy,
                                                            self._hub.hours.prices,
                                                            self._hub.hours.prices_tomorrow)
        await self.async_check_states()

    def cancel(self):
        self._cancel()
        self.schedule_created = False

    async def async_cancel(self):
        await self.hub.state_machine.async_add_executor_job(self._cancel)
        self.schedule_created = False

    def check_states(self):
        if not self.scheduler_active and self.schedule_created:
            self.cancel()
        elif self._hub.chargecontroller.status_string is ChargeControllerStates.Done.name:
            self.cancel()

    @property
    def non_hours(self) -> list:
        return self.model.non_hours

    @property
    def caution_hours(self) -> dict:
        """dynamic caution hours"""
        return self.model.caution_hours




class Schedule:
    start_date: date
    end_date: date
    start_time: time
    end_time: time
    charge_amount: float
    weekdays: int


class Schedules:
    #list of schedule with id
    pass

class SchedulerCommands:
    def __init__(self) -> None:
        pass

    def create(self) -> int:
        pass

    def edit(self, scheduler_id: int) -> bool:
        pass

    def delete(self, scheduler_id: int) -> bool:
        pass

    def pause(self, scheduler_id: int) -> bool:
        pass

    def resume(self, scheduler_id: int) -> bool:
        pass


# import pytest
# from ..hub.hub import Hub
# from ..hub.hub_options import HubOptions, Price
# from ..services.locale.Locale import LOCALE_SE_GOTHENBURG
# from ..models.hourselection.cautionhourtype import CautionHourType

# OPTIONS_REGULAR = HubOptions(
#     price=Price(
#         price_aware=True,
#         cautionhour_type=CautionHourType.SUAVE.value
#     ),
#     peaqev_lite=False,
#     startpeaks={1:2, 2:2, 3:1.8, 4:1.5, 5:1.5, 6:1.5, 7:1.5, 8:1.5, 9:1.5, 10:1.8, 11:2, 12:2}
# )
# OPTIONS_REGULAR.locale = LOCALE_SE_GOTHENBURG

# MOCK_CHARGER_OBJ = ""
# MOCK_STATE_MACHINE = ""

# def test_hub_init():
#     h = Hub(options=OPTIONS_REGULAR,domain="test", state_machine=MOCK_STATE_MACHINE, chargerobj = MOCK_CHARGER_OBJ, is_test=True)
#     assert h.domain == "test"
#     assert h.hours.price_aware is True
#     assert h.scheduler.active is False
    


from ....services.locale.querytypes.queryservice import QueryService
from ....models.locale.enums.querytype import QueryType
from ..time_pattern import TimePattern
from ....models.locale.enums.calendar_periods import CalendarPeriods
from ..locale_model import Locale_Type
from dataclasses import dataclass
from ..querytypes.querysets import QUERYSETS
from ..querytypes.querytypes import QUERYTYPES
from ..querytypes.const import (
    QUERYTYPE_NEVER
)

@dataclass(frozen=True)
class Default(Locale_Type):
    observed_peak = QueryType.Max
    charged_peak = QueryType.Max
    query_model = QUERYTYPES[QueryType.Max]


@dataclass(frozen=True)
class NoPeak(Locale_Type):
    observed_peak = QueryType.Max
    charged_peak = QueryType.Max
    query_model = QUERYTYPES[QueryType.Max]
    query_model.set_query_service(QueryService(QUERYSETS[QUERYTYPE_NEVER]))
    free_charge_pattern = TimePattern([
        {
            CalendarPeriods.Month: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
            CalendarPeriods.Hour: [0, 1, 2, 3, 4, 5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23]
        }
    ])


# from ...hub.hub_options import HubOptions
# from .chargecontroller import ChargeController
# from .chargecontroller_lite import ChargeControllerLite

# class ChargeControllerFactory:
#     @staticmethod
#     def create(options: HubOptions):
#         if options.peaqev_lite:
#             return ChargeControllerLite(options)
#         return ChargeController(options)